/***************************************************************************
 *                                                                         *
 *  MAXIMUS-CBCS Source Code, Version 1.02                                 *
 *  Copyright 1989, 1990 by Scott J. Dudley.  All rights reserved.         *
 *                                                                         *
 *  File-area routines (overlay 1)                                         *
 *                                                                         *
 *  For complete details of the licensing restrictions, please refer to    *
 *  the licence agreement, which is published in its entirety in MAX.C     *
 *  and LICENCE.MAX.                                                       *
 *                                                                         *
 *  USE OF THIS FILE IS SUBJECT TO THE RESTRICTIONS CONTAINED IN THE       *
 *  MAXIMUS-CBCS LICENSING AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF THIS *
 *  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES, OR IF YOU DO NOT HAVE   *
 *  THESE FILES, YOU SHOULD IMMEDIATELY CONTACT THE AUTHOR AT ONE OF THE   *
 *  ADDRESSES LISTED BELOW.  IN NO EVENT SHOULD YOU PROCEED TO USE THIS    *
 *  FILE WITHOUT HAVING ACCEPTED THE TERMS OF THE MAXIMUS-CBCS LICENSING   *
 *  AGREEMENT, OR SUCH OTHER AGREEMENT AS YOU ARE ABLE TO REACH WITH THE   *
 *  AUTHOR.                                                                *
 *                                                                         *
 *  You can contact the author at one of the address listed below:         *
 *                                                                         *
 *  Scott Dudley           FidoNet  1:249/106                              *
 *  777 Downing St.        IMEXnet  89:483/202                             *
 *  Kingston, Ont.         Internet f106.n249.z1.fidonet.org               *
 *  Canada - K7M 5N3       BBS      (613) 389-8315 - HST/14.4K, 24hrs      *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <mem.h>
#include <errno.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include "zmodem.h"
#include "prog.h"
#include "ffind.h"
#include "alc.h"
#include "max.h"
#include "max_msg.h"
#include "arc.h"

/* Some prototypes for near functions.  Declaring these routines as        *
 * "static <type> near pascal" makes them slightly faster to call, but     *
 * makes them accessible only within this module.  Hence, only the         *
 * timing-critical functions are marked as near.                           */

static long near pascal XferTime(int protocol,long bytes);
static int near pascal IsInFilesBbs(char *filename);
static int near pascal Match_Wildcard(char *filename,char *patstem,char *patext);
static int near pascal Match_Wildcard_Sub(char *filename,char *pattern);

static int near pascal IsValidFileArea(struct _area *a,int areanum);


int pascal File_Area(void)
{
  int savearea;

  savearea=usr.files;

  while (1)
  {
    barricade_ok=FALSE;
    FixMenuPath();
    File_Area_Menu();

    if (! IsValidFileArea(NULL,usr.files))
    {
      logit(denied_access,s(deny_file),Area_Name(usr.files),areatag);

      usr.files=savearea;

      if (*PRM(areanotexist))
        Display_File(0,NULL,PRM(areanotexist));
      else Puts(areadoesntexist);

      continue;
    }

    logit(s(log_farea),Area_Name(usr.files),areatag);

    if (*area.filebar)
      Barricade(FALSE);
    else
    {
      barricade_priv=usr.priv;
      barricade_ok=TRUE;
    }

    if (! *area.filebar || barricade_ok)
      break;
    else if (barricade_priv==0x1234)
    {
      barricade_priv=usr.priv;

      if (*PRM(areanotexist))
        Display_File(0,NULL,PRM(areanotexist));
      else Puts(areadoesntexist);
    }
  }

  if (Save_Directory(area.filepath)==-1)
  {
    logit(mem_ndir);
    Puts(mem_ndir);
    quit(ERROR_recycle);
  }

  entry=ENTRY_nochecke;

  return barricade_priv;
}


void pascal File_Area_Menu(void)
{
  struct _area carea;
  struct _aidx aidx;

  int first;

  long y;

  char temp[PATHLEN],
       string[PATHLEN],
       wrapped;

  WhiteN();

  first=TRUE;

  while (1)
  {
    barricade_ok=FALSE;  /* Reset barricade counter on A)rea cmd */

    Puts(WHITE);
    sprintf(string,s(file_prmpt),PRM(achg_keys)[0],PRM(achg_keys)[1],
            PRM(achg_keys)[2]);

    OpusGets(temp,string);
    strupr(temp);

    if (*temp=='\'' || *temp=='`')
      Display_File(0,NULL,ss,PRM(misc_path),quotes_misunderstood);
    else if (*temp=='#')              /* Maybe the user misunderstood? */
      Display_File(0,NULL,ss,PRM(misc_path),numsign_misunderstood);
    else if (*temp==PRM(achg_keys)[1] || *temp==']' || *temp=='>' ||
             *temp=='+')
    {
      wrapped=FALSE;
      strcpy(linebuf,temp+1);

      /* If we can't find our area, then return...  This call will         *
       * position the `areasidx' file pointer correctly...                 */

      if (Index_of_Area(usr.files)==-1L)
        return;

      /* Skip over the current area */
      lseek(areasidx,(long)sizeof(struct _aidx),SEEK_CUR);

      while (1)
      {
        /* Loop while there are still index entries */
        for (y=1;y > 0;)
        {
          /* Now get the next index entry... */

          y=read(areasidx,(char *)&aidx,sizeof(struct _aidx));

          lseek(areasfile,aidx.offset,SEEK_SET);
          read(areasfile,(char *)&carea,sizeof(struct _area));

          if (aidx.area <= prm.high_filearea && IsValidFileArea(&carea,0) &&
              ExtBarricadeOK(&carea,FALSE,NULL))
          {
            usr.files=carea.areano;          /* Found the right area... */
            return;
          }
        }

        if (!wrapped)   /* If we wrap around... */
        {
          wrapped=TRUE;
          lseek(areasidx,0L,SEEK_SET);
        }
        else return;
      }
    }
    else if (*temp==PRM(achg_keys)[0] || *temp=='[' || *temp=='<' ||
             *temp=='-')
    {
      wrapped=FALSE;
      strcpy(linebuf,temp+1);

      /* If we can't find our area, then return...  This call will         *
       * position the `areasidx' file pointer correctly...                 */

      if (Index_of_Area(usr.files)==-1L)
        return;

      while (1)
      {
        /* Loop while there are still index entries */

        for (y=1;y != 0;)
        {
          /* Now get the next index entry... */

          y=lseek(areasidx,-(long)sizeof(struct _aidx),SEEK_CUR);

          if (y < 0)
            break;

          read(areasidx,(char *)&aidx,sizeof(struct _aidx));
          lseek(areasidx,-(long)sizeof(struct _aidx),SEEK_CUR);

          lseek(areasfile,aidx.offset,SEEK_SET);
          read(areasfile,(char *)&carea,sizeof(struct _area));

          if (aidx.area <= prm.high_filearea && IsValidFileArea(&carea,0) &&
              ExtBarricadeOK(&carea,FALSE,NULL))
          {
            usr.files=carea.areano;          /* Found the right area... */
            return;
          }
        }

        if (!wrapped)   /* If we wrap around... */
        {
          wrapped=TRUE;
          lseek(areasidx,0L,SEEK_END);
        }
        else return;
      }
    }
    else if (*temp==PRM(achg_keys)[2] || *temp=='?' || *temp=='/')
    {
      strcpy(linebuf,temp+1);
      List_File_Areas();
    }
    else if (! *temp ||
        (temp[0] >= '0' && temp[0] <= '9') ||
        (temp[0] >= 'A' && temp[0] <= 'Z'))
    {
      if (*temp)
      {
        usr.files=Set_Area_Name(temp);
        return;
      }
      else if (first)
        List_File_Areas();
      else return;
    }
    else Printf(dontunderstand,temp[0]);

    first=FALSE;
  }
}




void pascal File_Download(char *mname)
{
  struct _css css;
  char temp[BUFLEN];
  int protocol;         /* Protocol to use for download */

  if (local)
  {
    Puts(err999_1);
    OpusGets(temp,err999_2);
  }
  else
  {
    if (baud < prm.class[class].min_file_baud)
    {
      Display_File(0,NULL,PRM(xferbaud));
      Printf(baudtoolowforxfer,prm.class[class].min_file_baud);
      Press_ENTER();
    }
    else
    {
      if (File_Get_Protocol(&protocol)==-1)
        return;

      if (! Alloc_Filenames_Buffer())
        return;

      ChatSaveStatus(&css);
      ChatSetStatus(FALSE,s(cs_file_xfer));

      File_Get_Names(protocol,FALSE);

      if (fnames) /* At least one file to send */
        File_Send_Files(protocol,mname);
      else Puts(xferaborted);

      Free_Filenames_Buffer();

      ChatRestoreStatus(&css);
    }
  }
}



void pascal File_Upload(char *mname)
{
  struct _css css;
  char temp[BUFLEN];
  int protocol;         /* Protocol to use for upload */
  long b_free;

  if (*area.uppath && direxist(area.uppath))
  {
    if (local)
    {
      Puts(err999_1);
      OpusGets(temp,err999_2);
      baud=9600L;
    }
    else
    {
      if (baud < prm.class[class].min_file_baud)
      {
        Display_File(0,NULL,PRM(xferbaud));
        Printf(baudtoolowforxfer,prm.class[class].min_file_baud);
        Press_ENTER();
      }
      else
      {
        b_free=zfree(area.uppath);
          
        if (prm.k_free && b_free < (long)prm.k_free*1000L)
        {
          Display_File(0,NULL,PRM(no_space));
          return;
        }

        Printf(s(bytes_for_ul),b_free-(long)prm.k_free*1000L);

        if (File_Get_Protocol(&protocol)==-1)
          return;

        if (! Alloc_Filenames_Buffer())
          return;

        ChatSaveStatus(&css);
        ChatSetStatus(FALSE,s(cs_file_xfer));

        if (protocol==PROTOCOL_xmodem || protocol==PROTOCOL_xmodem1k ||
            (protocol >= 0 && !(prm.protoflag[protocol] & XTERNBATCH)))
          File_Get_Names(protocol,TRUE);
        else fnames=1;

        if (fnames) /* At least one file to get */
          File_Get_Files(protocol,mname);
        else Puts(xferaborted);

        Free_Filenames_Buffer();

        ChatRestoreStatus(&css);
      }
    }
  }
  else
  {
    sprintf(temp,"%sNOUPLOAD.?BS",area.filepath);

    if (fexist(temp))
    {
      /* Chop off extension... */
      *strrchr(temp,'.')='\0';

      /* ...and display */
      Display_File(0,NULL,temp);
    }
    else
    {
      Puts(s(noupl));
      Press_ENTER();
    }
  }
}




int pascal File_Get_Protocol(int *protocol)
{
  char temp[PATHLEN];

  int ch,
      x;

  *protocol=PROTOCOL_none;

  if (! *linebuf)
  {
    if (*PRM(proto_dump))
      Display_File(0,NULL,PRM(proto_dump));
    else
    {
      Puts(s(pick_proto));

      if (!no_zmodem)
        Puts(s(proto_zmodem));

      for (x=0;x < MAX_EXTERNP;x++)
      {
        if (*(PRM(protoname[x])))
        {
          if (prm.protoname[x]==prm.protocols[x])
            getword(Last_Leaf(PRM(protocols[x])),temp,".",1);
          else strcpy(temp,PRM(protoname[x]));

          Printf("  %c)%s\n",toupper(temp[0]),&temp[1]);
        }
      }

      Puts(s(proto_quit));
    }
  }

  ch=KeyGetRNP(select_p);

  switch (ch=toupper(ch))
  {
    case 'X':
      *protocol=PROTOCOL_xmodem;
      break;

    case '1':
      *protocol=PROTOCOL_xmodem1k;
      break;

    case 'S':
      *protocol=PROTOCOL_sealink;
      break;

    case 'T':
      *protocol=PROTOCOL_telink;
      break;

    case '\r':
    case '\x00':
    case 'Q':
      return -1;

    case 'Z':
      if (!no_zmodem)
      {
        *protocol=PROTOCOL_zmodem;
        break;
      }
      /* else fall-through, pretend internal Zmodem doesn't exist */

    default:
      for (x=0;x < MAX_EXTERNP;x++)
      {
        if (prm.protoname[x]==prm.protocols[x])
          getword(Last_Leaf(PRM(protocols[x])),temp,".",1);
        else *temp=*PRM(protoname[x]);

        if (ch==toupper(*temp))
        {
          *protocol=x;
          break;
        }
      }

      if (x==MAX_EXTERNP)  /* Unknown response */
      {
        mknl(linebuf);
        Printf(s(i_dont_get_it),ch);
        return -1;
      }
      break;
  }

  return 0;
}



void pascal File_Get_Names(int protocol,int upload)
{
  FFIND *ff;
  char expanded[MAXEXPAND];

  char temp[255],
       *p;

  int ret,
      x;

  WhiteN();

  if (upload)
    OpusGetsLL(temp,254,s(file_ul));
  else OpusGetsLL(temp,254,s(file_dl));

  fnames=0;
  p=strtok(temp,dl_delim);

  while (1)
  {
    if (!p)
      break;

    Strip_Path(p);


    ff=FindOpen(p,0);

    if (ff==NULL)   /* Wasn't found, no wildcard matching. */
    {
      expanded[fnames]=FALSE;
      strncpy(filenames[fnames],p,13);
      filenames[fnames++][12]='\0';
    }
    else
    {
      for (ret=0;ret==0;ret=FindNext(ff))
      {
        if (fnames >= MAXEXPAND)
          break;

        if (eqstri(ff->ff_name,p))
          expanded[fnames]=FALSE;
        else expanded[fnames]=TRUE;

        strcpy(filenames[fnames],ff->ff_name);
        filenames[fnames++][12]='\0';
      }
    }

    FindClose(ff);

    p=strtok(NULL,dl_delim); /* Get the next match */
  }

  if (fnames > 1 && (protocol==PROTOCOL_xmodem ||
                     protocol==PROTOCOL_xmodem1k ||
                     (protocol >= 0 && !(prm.protoflag[protocol] & XTERNBATCH))))
  {
    Puts(s(file_only1));
    fnames=0;
    return;
  }

  if (! upload) /* Only verify FILES.BBS entries if downloading... */
  {
    Puts(CLS WHITE);

    for (x=0;x < fnames;x++)
    {
      if (IsInFilesBbs(filenames[x]))
      {
        if (! fexist(filenames[x]))
        {
          if (!expanded[x])
            File_IsOffline(filenames[x]);

          Remove_File_From_List(x);
          x--;
        }
        /* else DL is okay! */
      }
      else
      {
        if (GEPriv(usr.priv,prm.dlall_priv)) /* Priv. needed to download file not in FILES.BBS */
        {
          if (! fexist(filenames[x]))
          {
            if (!expanded[x])
              File_NotExist(filenames[x]);

            Remove_File_From_List(x);
            x--;
          }
          /* else DL is okay! */
        }
        else
        {
          if (!expanded[x])
            File_NotExist(filenames[x]);

          Remove_File_From_List(x);
          x--;
        }
      }
    }

    Putc('\n');
  }
  else
  {
    /* Else we're uploading, so check for a valid filename */

    getword(filenames[0],temp,dot,2);

    if (strlen(filenames[0]) > 12 || strlen(temp) > 3 ||
        strchr(filenames[0],' ') || strchr(filenames[0],'|') ||
        strchr(filenames[0],'>') || strchr(filenames[0],'>'))
    {
      Display_File(0,NULL,PRM(fname_format));
      *filenames[0]='\0';
      fnames=0;
    }
  }
}


void pascal File_Send_Files(int protocol,char *mname)
{
  char cmd[PATHLEN],
       temp[PATHLEN],
       temp2[PATHLEN],
       xferstatus;

  long bytes;

  int x,y;

  for (x=0,bytes=0;x < fnames;x++)
    bytes += fsize(filenames[x]);

  Puts(s(fnam));

  for (x=0;x < fnames;x++)
  {
    /* Word-wrap the filenames */
    
    if (current_col+strlen(filenames[x]) > usr.width)
      Printf("\n\x19 %c",strlen(s(fnam)));
    
    Printf("%s ",strupr(filenames[x]));
  }

  Printf(s(fsiz),bytes,(bytes/128L)+1);

  Printf(s(ftim),
         (unsigned int)(XferTime(PROTOCOL_zmodem,bytes)/60L),
         (unsigned int)(XferTime(PROTOCOL_zmodem,bytes)%60L),
         (unsigned int)(XferTime(PROTOCOL_xmodem,bytes)/60L),
         (unsigned int)(XferTime(PROTOCOL_xmodem,bytes)%60L),
         (unsigned int)(XferTime(PROTOCOL_sealink,bytes)/60L),
         (unsigned int)(XferTime(PROTOCOL_sealink,bytes)%60L));

  Printf(s(fmodd),Protocol_Name(protocol,temp));

  if (usr.dnldl+(unsigned int)(bytes/1000L) > prm.class[class].max_dl)
  {
    logit(s(log_dllim));
    Puts(s(exc_lim));
    Press_ENTER();
  }
  else if (usr.dnld+(unsigned int)(bytes/1000L) > prm.ratio_threshold &&
           prm.class[class].ratio &&
           (usr.upld==0 ? (usr.dnld+(unsigned)(bytes/1000L)) :
                          (usr.dnld+(unsigned)(bytes/1000L))/usr.upld) >
           prm.class[class].ratio)
  {
    logit(s(log_exc_ratio));

    Printf(s(exc_ratio),prm.class[class].ratio);

    Press_ENTER();
  }
  else if (time(NULL)+XferTime(protocol,bytes) > timeoff)
  {
    Puts(s(exc_limit));
    Press_ENTER();
  }
  else
  {
    Puts(s(start_receive));

    if (protocol==PROTOCOL_xmodem || protocol==PROTOCOL_xmodem1k)
      Printf("%s ",filenames[0]);

    Puts(s(end_receive));

    /* Force log messages to appear on-screen */
    in_file_xfer=TRUE;
    last_bps=0;

    logit(s(log_sending_to),usr.name,baud);

    /* Turn off ^S/^Q flow control and ^C/^K checking, so it doesn't
       screw up the transfers!                                       */

    Mdm_flush();

    Mdm_Flow_Off();

#ifdef INTERNAL_PROTOCOLS

    if ((Txbuf=Secbuf=zalloc()) != NULL)
    {
#endif

      switch (protocol)
      {
#ifdef INTERNAL_PROTOCOLS
        case PROTOCOL_xmodem:
        case PROTOCOL_telink:
        case PROTOCOL_xmodem1k:
        case PROTOCOL_sealink:
          switch (protocol)
          {
            case PROTOCOL_xmodem:
              protocol_letter='X';
              break;

            case PROTOCOL_telink:
              protocol_letter='T';
              break;

            case PROTOCOL_xmodem1k:
              protocol_letter='Y';
              break;

            case PROTOCOL_sealink:
              protocol_letter='S';
              break;
          }

        case PROTOCOL_zmodem:
          xferstatus=TRUE;

          Lputc('\n');

          for (y=0;y < fnames;y++)
          {
            /* Have filenames in this format:  d:\path\FILENAME.EXT */

            strcpy(temp,area.filepath);
            strlwr(temp);
            strcat(temp,strupr(filenames[y]));

            if (protocol==PROTOCOL_zmodem)
              x=Send_Zmodem(temp,NULL,0,FALSE);
            else
            {
              if (protocol==PROTOCOL_telink)
                x=xfermdm7(temp);
              else x=TRUE;

              if (x)  /* If good Telink filename, or X/Y/SEAlink */
                x=send_file(temp,protocol_letter);
            }

            if (x)  /* This file sent OK! */
            {
              if ((protocol==PROTOCOL_zmodem && x != SPEC_COND) ||
                  protocol != PROTOCOL_zmodem)
              {
                /* If we're using Zmodem, don't debit downloads if *
                 * we got a ZSKIP header!                          */

                usr.dnld += (unsigned int)(fsize(filenames[y])/1000L);
                usr.dnldl += (unsigned int)(fsize(filenames[y])/1000L);
                bstats.total_dl += (long)(fsize(filenames[y])/1000L);
              }
            }
            else
            {
              Delay(50);

              xferstatus=FALSE;
              logit(lxferaborted);
              Puts(xferaborted);
              break;
            }
          }

          if (xferstatus) /* End batch transfer! */
          {
            if (protocol==PROTOCOL_zmodem)
              Send_Zmodem(NULL,NULL,END_BATCH,FALSE);
            else send_file(NULL,protocol_letter);

            Delay(50);

            Puts(xfercomplete);

            if (last_bps)
              Printf(cps_rating,(int)(last_bps/10L),(int)(last_bps*100L/baud));
            else Putc('\n');
          }
          break;

#else
        case PROTOCOL_zmodem:     /* EXTERNAL X/Y/ZMODEM PROTOCOLS! */
        case PROTOCOL_xmodem:
        case PROTOCOL_telink:
        case PROTOCOL_xmodem1k:
        case PROTOCOL_sealink:

          if (protocol==PROTOCOL_sealink)
            sprintf(cmd,"command /c CLINK /P%u /B%ld T",port+1,baud);
          else
            sprintf(cmd,"command /c DSZ port %u speed %ld s%s",port+1,
                    baud,
                    protocol==PROTOCOL_xmodem || protocol==PROTOCOL_telink ? "x" :
                    (protocol==PROTOCOL_xmodem1k ? "x -k" :
                     "z"));

          for (x=0;x < fnames;x++)
            sprintf(&cmd[strlen(cmd)]," %s",filenames[x]);

          Lprintf("Cmd: `%s'\n",cmd);

          if (Outside(NULL,NULL,OUTSIDE_run,cmd,FALSE,CTL_none,RESTART_menu,NULL)==0)
          {
            for (x=0;x < fnames;x++)
            {
              switch(protocol)
              {
                case PROTOCOL_xmodem:
                  logit("=DL-%c %s%s",'X',area.filepath,fancy_str(filenames[x]));
                  break;

                case PROTOCOL_xmodem1k:
                  logit("=DL-%c %s%s",'Y',area.filepath,fancy_str(filenames[x]));
                  break;

                case PROTOCOL_zmodem:
                  logit("=DL-%c %s%s",'Z',area.filepath,fancy_str(filenames[x]));
                  break;

                case PROTOCOL_sealink:
                  logit("=DL-%c %s%s",'S',area.filepath,fancy_str(filenames[x]));
                  break;
              }
            }

            usr.dnld += (unsigned int)(bytes/1000L);
            usr.dnldl += (unsigned int)(bytes/1000L);
            bstats.total_dl += (long)(bytes/1000L);

            Puts(xfercomplete);
            Putc('\n');
          }
          else
          {
            logit(lxferaborted);
            Puts(xferaborted);
          }
          break;
#endif

        default:
          protocol_letter=*Protocol_Name(protocol,temp);

          getword(PRM(protocols[protocol]),temp,".",1);
          strcat(temp,".Ctl");
          Add_Task_Number(temp,temp2);

          if (prm.protoflag[protocol] & XTERNEXIT)
            sprintf(temp,"%d ",prm.protoexit);
          else *temp='\0';

          sprintf(cmd,xtrn_caller,temp,PRM(protocols[protocol]),temp2);

          Outside(NULL,NULL,(prm.protoflag[protocol] & XTERNEXIT) ?
                  OUTSIDE_errorlevel : OUTSIDE_run,cmd,FALSE,CTL_download,
                  RESTART_menu,mname);
          break;
      }

#ifdef INTERNAL_PROTOCOLS
      free(Txbuf);
      Txbuf=Secbuf=NULL;

    }
#endif

    mdm_dump(DUMP_input);
    Clear_KBuffer();

    in_file_xfer=FALSE;

    /* And turn flow control back on! */
    Mdm_Flow_On();
  }
}




char * pascal Protocol_Name(int protocol,char *temp)
{
  switch(protocol)
  {
    case PROTOCOL_none:     /* ?!? */
      return blank_str;

    case PROTOCOL_xmodem:
      return (s(prot_xmod));

    case PROTOCOL_xmodem1k:
      return (s(prot_xm1k));

    case PROTOCOL_zmodem:
      return (s(prot_zmod));

    case PROTOCOL_sealink:
      return (s(prot_seal));

    case PROTOCOL_telink:
      return (s(prot_tel));

    default:              /* External protocol */
      if (prm.protoname[protocol]==prm.protocols[protocol])
        getword(Last_Leaf(PRM(protocols[protocol])),temp,".",1);
      else strcpy(temp,PRM(protoname[protocol]));

      return temp;
  }
}



static long near pascal XferTime(int protocol,long bytes)
{
  long cps,
       seconds;

  cps=baud/10;
  seconds=bytes/cps;

  switch(protocol)
  {
    case PROTOCOL_none:     /* ?!? */
      return 0L;

    case PROTOCOL_xmodem: /* 72% efficiency: 100/.72==138 */
    case PROTOCOL_telink:
      seconds=(seconds*138)/100;
      break;

    case PROTOCOL_xmodem1k: /* 91% efficiency: 100/.91==110 */
      seconds=(seconds*110)/100;
      break;

    case PROTOCOL_sealink:/* 93% efficiency: 100/.93==107 */
      seconds=(seconds*107)/100;
      break;

    case PROTOCOL_zmodem: /* 96% efficiency: 100/.96==104 */
      seconds=(seconds*104)/100;
      break;

    default:              /* External protocol */
      seconds=(seconds*111)/100;  /* Avg. 90% efficiency: 100/.90==111 */
      break;
  }

  return seconds;
}



void pascal File_IsOffline(char *filename)
{
  Printf(s(file_offl),strupr(filename));
  mknl(filename);
}


void pascal File_NotExist(char *filename)
{
  Printf(iseenoxhere,strupr(filename));
  mknl(filename);
}



static int near pascal IsInFilesBbs(char *filename)
{
  FILE *filesbbs;

  long offset;

  int ret;

  char temp[255],
       *filebufr,
       patstem[13],
       patext[13],
       *p;

  getword(filename,patstem,".\n",1);
  getword(filename,patext,".\n",2);

  ret=FALSE;
  filebufr=NULL;

  if ((filesbbs=shfopen(*area.filesbbs ? area.filesbbs : files_bbs,fopen_read,O_RDONLY))==NULL)
    logit(cantopen,*area.filesbbs ? area.filesbbs : files_bbs);
  else
  {
    if ((filebufr=malloc(FILEBUFSIZ)) != NULL)
      setvbuf(filesbbs,filebufr,_IOFBF,FILEBUFSIZ);

    while (1)
    {
      offset=ftell(filesbbs);

      fgets(temp,255,filesbbs);

      if (feof(filesbbs))
      {
        fclose(filesbbs);
        ret=FALSE;
        goto DoRet;
      }
      else
      {
ReLoop: switch(*temp)
        {
          case '\x10':        /* Priv. command! */
            fseek(filesbbs,offset+1L,SEEK_SET);

            switch(Parse_Priv(filesbbs))
            {
              case SKIP_none: /* Grab line again */
                fgets(temp,255,filesbbs);
                offset=ftell(filesbbs);
                goto ReLoop;  /* In case there are any more priv commands... */

              case SKIP_file:
                goto End_IsInFilesBbs;

              case SKIP_line:             /* Go back for another line */
                Skip_Line(filesbbs);
                continue;
            }
            break;

          case '@':
            if (! GEPriv(usr.priv,ASSTSYSOP))
            {
              fclose(filesbbs);
              ret=FALSE;
              goto DoRet;
            }
            else memmove(temp,&temp[1],strlen(&temp[1])+1); /* Get the @ outta the way */
            /* fall-through */

          default:
            if (*temp > ' ' && *temp != '-')
            {
              temp[12]='\0';

              p=temp;

              /* Get everything up 'till the first space, the fast way. */

              while (1)
              {
                if (*p==' ' || *p=='\n')
                {
                  *p='\0';
                  break;
                }
                else if (! *p++)
                  break;
              }

              if (Match_Wildcard(temp,patstem,patext))
              {
                fclose(filesbbs);
                ret=TRUE;
                goto DoRet;
              }
            }
            break;
        }
      }
    }

End_IsInFilesBbs:

    fclose(filesbbs);
  }

DoRet:

  if (filebufr)
    free(filebufr);

  return ret;
}



static int near pascal Match_Wildcard(char *filename,char *patstem,char *patext)
{
  char *filestem,
       *fileext;

  if ((filestem=strtok(filename,".\n"))==NULL)
    filestem=blank_str;

  if ((fileext=strtok(NULL,".\n"))==NULL)
    fileext=blank_str;

  if (Match_Wildcard_Sub(filestem,patstem) &&
      Match_Wildcard_Sub(fileext,patext))
    return TRUE;
  else return FALSE;
}

static int near pascal Match_Wildcard_Sub(char *filename,char *pattern)
{
  static int x,
             slf,
             slp;

  slf=strlen(filename);
  slp=strlen(pattern);

  for (x=0;x < slp;x++)
  {
    if (filename[x] != '?')
    {
      if (filename[x]=='*')
        return TRUE;
      else if (toupper(filename[x]) != toupper(pattern[x]) || x >= slf)
        return FALSE;
    }
  }

  if (x < slf && filename[0] != '*')
    return FALSE;
  else return TRUE;
}






void pascal File_Area_Tag(void)
{
  strcpy(areatag,Last_Leaf(area.filepath));
}




void pascal File_Type(void)
{
  FILE *typefile;

  char filename[PATHLEN],
       filebufr[FILEBUFSIZ],
       nonstop;

  unsigned char buf[PATHLEN];

  int ch,
      c2,
      x,y,z;

  WhiteN();

  OpusGets(filename,s(type_which));

  WhiteN();

  Strip_Path(filename); /* Make sure user doesn't specify path or device */

  if (! isnl(filename))
  {
    if (File_Okay(filename))
    {
      if ((typefile=shfopen(filename,fopen_readb,O_RDONLY | O_BINARY))==NULL)
        logit(cantopen,filename);
      else
      {
        setvbuf(typefile,filebufr,_IOFBF,FILEBUFSIZ);

        /* Get a line into buf, and check for binary characters */
        z=fread(buf,1,PATHLEN-1,typefile);

        /* If it's an ARC, EXE or ZIP header */
        if (*buf=='\x1a' || strncmp(buf,"MZ",2)==0 || strncmp(buf,"PK",2)==0 ||
            stristr(filename,dot_zip) || stristr(filename,dot_pak) ||
            stristr(filename,dot_lzh) || stristr(filename,dot_arc))
          x=0;
        else
        {
          /* Break if anything is below ASCII 32, with the exception of
             returns, linefeeds, and ESCape characters (for ANSI!)      */
          for (x=y=0;x < min(z,PATHLEN-1);x++)
            if (buf[x] < ' ' && buf[x] != '\r' && buf[x] != '\n' &&
                buf[x] != '\x1b')
              y++;
        }

        if (x==0 || y >= 15) /* If we found a MZ or many control characters */
        {
          Puts(s(type_for_text));
          Press_ENTER();
        }
        else
        {
          logit(s(log_disp),area.filepath,filename);

          nonstop=FALSE;
          fseek(typefile,0L,SEEK_SET);

          display_line=display_col=1;

          while (1)
          {
            ch=fgetc(typefile);

            if (ch != EOF && ch != '\x1b')
            {
              if (ch=='\r')
              {
                if ((c2=fgetc(typefile)) != '\n')
                  ch='\n';
                
                ungetc(c2,typefile);
              }
              
              Putc(ch);

              if (display_col==1 && MoreYnBreak(&nonstop,NULL))
                break;
            }
            else break;
          }
        }

        fclose(typefile);
      }
    }
  }
}



int pascal File_Okay(char *filename)
{
  int okay;
  FFIND *ff;

  okay=TRUE;

  if (IsInFilesBbs(filename))
  {
    ff=FindOpen(filename,0);

    if (!ff || !eqstri(filename,ff->ff_name))
    {
      File_IsOffline(filename);
      okay=FALSE;
    }

    FindClose(ff);
  }
  else
  {
    if (GEPriv(usr.priv,prm.dlall_priv)) /* Priv. needed to download file not in FILES.BBS */
    {
      ff=FindOpen(filename,0);

      if (!ff || !eqstri(filename,ff->ff_name))
      {
        File_NotExist(filename);
        okay=FALSE;
      }

      FindClose(ff);
    }
    else
    {
      File_NotExist(filename);
      okay=FALSE;
    }
  }

  return okay;
}


void pascal Dont_Use_Wildcards(char *name)
{
  Printf(s(no_wildcard),name);

  Puts(s(no_wc1));
  Puts(s(no_wc2));
  Press_ENTER();
}



int pascal Header_File(int entry)
{
  int usrpriv,
      nochke;
    
  nochke=FALSE;
  
  if (entry==ENTRY_nochecke)
  {
    entry=ENTRY_nocheck;
    nochke=TRUE;
  }

  if (inmsgarea)
  {
    Fix_Lastread();
    inmsgarea=FALSE;
  }

  if (entry==ENTRY_true || entry==ENTRY_noheader || entry==ENTRY_nocheck)
  {
    if (entry != ENTRY_nocheck)
    {
      if (!barricade_ok)
      {
        Undo_Barricade();

        if (! IsValidFileArea(NULL,usr.files))   /* Read SYSTEMxx and see if it's valid */
          usrpriv=File_Area();

        if (*area.filebar)
        {
          Barricade(FALSE);

          if (barricade_ok)
            usrpriv=barricade_priv;
          else usrpriv=File_Area();
        }
        else
        {
          usrpriv=usr.priv;
          barricade_ok=TRUE;
        }

        Redo_Barricade(usrpriv);
      }
    }

    if (Save_Directory(area.filepath)==-1)
    {
      logit(mem_ndir);
      Puts(mem_ndir);
      return FALSE;
    }

    if (isnl(linebuf) && entry != ENTRY_noheader &&
        (entry != ENTRY_nocheck || nochke) && !restart_system)
      Puts(s(files_banner));

    infilearea=TRUE;
    last=MENU_file;

    if (! isnl(area.uppath) && ! direxist(area.uppath))
      logit(s(bad_ul_path),Area_Name(usr.files));
  }

  if (isnl(linebuf) && entry != ENTRY_noheader &&
      (entry != ENTRY_nocheck || nochke))
    Printf(file_name,Area_Name(usr.files),area.fileinfo);

  entry=ENTRY_false;

  restart_system=FALSE;
  return TRUE;
}



int pascal Last_File_Overlay(void)
{
  last=MENU_file;
  entry=ENTRY_false;

  if (inmsgarea)
    usr.files=usr.msg;

  if (! IsValidFileArea(NULL,usr.files))
  {
    You_Cant_Get_There();
    return -1;
  }
  else
  {
    IsValidFileArea(NULL,usr.files);

    if (Save_Directory(area.filepath)==-1)
    {
      You_Cant_Get_There();
      return -1;
    }
    else
    {
      sfree(msgs);

      infilearea=TRUE;
      inmsgarea=FALSE;
    }
  }

  return 0;
}



void pascal Strip_Path(char *filename)
{
  char *p,
       **p2;

  static char *bad_names[]={"CON","COM","AUX","PRN","LPT","LPT1","LPT2",
                            "COM1","COM2","GATE1","GATE2","..",NULL};

  while ((p=strrchr(filename,'\\')) != NULL ||
         (p=strrchr(filename,'/')) != NULL ||
         (p=strrchr(filename,':')) != NULL)
  {
    *p='\0';
    logit(s(supp_path),fancy_str(filename));
    memmove(filename,p+1,strlen(p+1)+1);
  }

  for (p2=bad_names;*p2;p2++)
  {
    if (eqstri(filename,*p2))
    {
      logit(s(udev),fancy_str(filename));
      mknl(filename);
      return;
    }
  }

  if (strlen(filename) > 12)
    filename[12]='\0';
}


void pascal Add_To_Upload_Log(char *path,char *fname,long bytes)
{
  FILE *ullog;

  time_t ct;

  struct _stamp stamp;

  int ulfile;

  char temp[PATHLEN],
       p1[50],
       p2[50];

  TmDate_to_OpusDate(localtime(((ct=time(NULL)),&ct)),&stamp);

  if (*PRM(ul_log))
  {
    if ((ullog=shfopen(PRM(ul_log),fopen_append,O_WRONLY | O_APPEND))==NULL)
      logit(cantopen,PRM(ul_log));
    else
    {
      strcpy(p1,Timestamp_Format(PRM(dateformat),&stamp,temp));
      strcpy(p2,Timestamp_Format(PRM(timeformat),&stamp,temp));


      fprintf(ullog,s(ullog_format),
              usr.name,path,fname,bytes,p1,p2);

      fclose(ullog);
    }
  }

  /* Now touch the file's date so it is set to right now. */

  sprintf(p1,ss,path,fname);

  if ((ulfile=shopen(p1,O_RDWR | O_BINARY)) != -1)
  {
    set_fdt(ulfile,(union stamp_combo *)&stamp);
    close(ulfile);
  }
}






char *zalloc(void)
{
  byte *sptr;

  if ((sptr=malloc(WAZOOMAX+16))==NULL)
    logit("!Z-MEMOVFL");

  return sptr;
}



int Alloc_Filenames_Buffer(void)
{
  int x,y;

  for (x=0;x < MAXEXPAND;x++)
    if ((filenames[x]=malloc(PATHLEN*sizeof(char)))==NULL)
    {
      logit("!MEM:nfa");

      for (y=0;y < x;y++)
        free(filenames[y]);

      return FALSE;
    }

  return TRUE;
}



void Free_Filenames_Buffer(void)
{
  int x;

  for (x=0;x < MAXEXPAND;x++)
  {
    free(filenames[x]);
    filenames[x]=NULL;
  }
}



void pascal Remove_File_From_List(int n)
{
  int k;
  char *p;

  /* Remove filename from list */

  p=filenames[n];
  *p='\0';

  for (k=n;k < fnames-1;k++)
    filenames[k]=filenames[k+1];

  filenames[k]=p;

  fnames--;
}





void pascal List_File_Areas(void)
{
  struct _aidx aidx;
  struct _area carea;

  int x;

  char nonstop,
       temp[PATHLEN];


  if (! isnl(PRM(file_area_list)))
    Display_File(0,NULL,PRM(file_area_list));
  else
  {
    nonstop=FALSE;

    Puts(CLS);

    display_line=display_col=1;
    infilearea=TRUE;

    Parse_Custom_Area_Listing(NULL,PRM(file_header),temp,TRUE);
    Puts(temp);

    lseek(areasidx,0L,SEEK_SET);

    while (1)
    {
      x=read(areasidx,(char *)&aidx,sizeof(struct _aidx));

      if (x)
      {
        lseek(areasfile,aidx.offset,SEEK_SET);
        read(areasfile,(char *)&carea,sizeof(struct _area));

        if (aidx.area <= prm.high_filearea && IsValidFileArea(&carea,0) &&
            ExtBarricadeOK(&carea,FALSE,NULL))
        {
          Parse_Custom_Area_Listing(&carea,PRM(file_format),temp,FALSE);
          Puts(temp);
        }

        if (brk_trapped || mdm_halt())
        {
          brk_trapped=0;
          mdm_dump(DUMP_all);
          break;
        }

        if (MoreYnBreak(&nonstop,CYAN))
          break;
      }
      else break;
    }
    
    Parse_Custom_Area_Listing(NULL,PRM(file_footer),temp,FALSE);
    Puts(temp);

    Putc('\n');
  }
}



void pascal File_Kill(void)
{
  char filename[PATHLEN],
       path[PATHLEN],
       temp[PATHLEN];

  WhiteN();

  OpusGets(filename,s(file_to_kill));

  Strip_Path(filename); /* Make sure user doesn't specify path or device */

  if (! isnl(filename))
  {
    if (fexist(filename))
    {
      strcpy(path,area.filepath);
      strcat(path,filename);

      sprintf(temp,s(delete_yn),filename);

      if (GetyNAnswer(temp,0)==YES)
      {
        if (unlink(path)==-1)
        {
          Printf(s(cant_unlink),path);
          Press_ENTER();
        }
        else Remove_Files_Entry(filename,NULL);
      }
      else
      {
        if (GetyNAnswer(s(lfbrfbe),0)==YES)
          Remove_Files_Entry(filename,NULL);
      }
    }
    else    /* File doesn't exist, but it's in the FILES.BBS... */
    {
      if (IsInFilesBbs(filename))
        Remove_Files_Entry(filename,NULL);
      else Printf(cantfind+1,filename);
    }
  }
}

#include "isvfa.c"

